#
# TIConvert v0.2
# By Michael Lee
# Released Febuary 24 2011
#


import sys
import time
import optparse
import AxeTokensDict

# 'from X import all' doesn't appear to be working...

r_TwoByteTokens = AxeTokensDict.r_TwoByteTokens
r_AxePush = AxeTokensDict.r_AxePush
r_AxePop = AxeTokensDict.r_AxePop
r_AxeBothP = AxeTokensDict.r_AxeBothP
r_AxeNoP = AxeTokensDict.r_AxeNoP
d_AxeTokensMain = AxeTokensDict.d_AxeTokensMain


### Defining metadata: ###
exe_name = "TIconvert.py"
program_name = "TI Tokenizer/Detokenizer"
program_version = "0.2"
release_date = "Feb 25 2011"
author_name = "Michael Lee"
python_version = "3.1"
axe_version = "0.5.0"

help_string = "{0} v{1}.  Released {2}.  Written by {3} using Python version {4}.  " \
              "Converts to Axe v{5}.  To view help, type 'TIconvert.py --help'"
help_string = help_string.format(program_name, 
                                 program_version, 
                                 release_date, 
                                 author_name, 
                                 python_version,
                                 axe_version)

### Defining data: ###
d_VariableTypes = {0:"a real number",
                   1:"a list",
                   2:"a matrix",
                   3:"an equation",
                   4:"a string",
                   5:"a program",
                   6:"a protected program",
                   7:"a picture",
                   8:"a GDB",
                   9:"unknown",
                   10:"an unknown equation",
                   11:"a 'new equation'",
                   12:"a complex number",
                   13:"a complex list",
                   14:"undefined",
                   15:"a window variable",
                   16:"a ZSto variable",
                   17:"a table range variable",
                   18:"a LCD variable",
                   19:"a backup",
                   20:"an application (app)",
                   21:"an appvar",
                   22:"a temporary program",
                   23:"a group"}

r_AllowedTypes = [5, 6, 21, 22]

hexstring = "0123456789ABCDEF"

signature = "2A2A54493833462A1A0A00"

is_logging_enabled = False

tab_length = 0

errorlist = []

### Defining functions ###
def Write(message, override = 0):
    """This function is identical to print(X).

    To increase compatibility between Python 2.X and 3.X, this
    function ensures that 'print' is used only once for easy replacement.
    If logging is enabled, this also copies the text to the log, and
    if verbose is disabled, won't print anything to the console."""
    if options.verbose or (override >= 2):
        print("\n" + message)
        if override == 2:
            errorlist.append(message)
    if is_logging_enabled:
        try:
            ## Logging will be re-added when I can use argparse and
            ## compile 3.2 to exe.
            log_file.write(message)
        except:
            pass
    return

def NumHexToStr(number):
    """Converts a decimal number into a hexadecimal string.

    >>> NumHexToStr(12)
    '0C' """
    temp1 = number // 16
    temp2 = number % 16

    result = hexstring[temp1] + hexstring[temp2]
    return result

def ReturnByte(offset, buffer):
    """ Reads a single-byte token from memory (string)

    Returns the integer equivalent"""
    temp1 = hexstring.index(buffer[offset * 2])
    temp2 = hexstring.index(buffer[offset * 2 + 1])
    return temp1 * 16 + temp2

def ReturnTwoBytes(offset, buffer):
    """ Reads a two-byte token from memory (string)

    Returns an integer equivalent"""
    return ReturnByte(offset + 1, buffer) * 256 + ReturnByte(offset, buffer)

def DoesFileExist(path):
    try:
        file = open(path, "a")
    except:
        Write("ERROR: " + path + " does not appear to exist - did you type it in correctly?")
        raise
        QuitProgram(error = 2)
    file.close()
    return

def QuitProgram(message="Terminating program...", error=0):
    """Terminates the program.

    Also tries closing any opened files."""
    Write("Trying to close files...")
    try:
        input_file.close()
        Write("Successfully closed input file")
    except:
        Write("Input file already closed.")

    try:
        output_file.close()
        Write("Successfully closed output file")
    except:
        Write("Output file already closed.")
        
    try:
        log_file.close()
        Write("Successfully closed log")
    except:
        pass

    if errorlist == []:
        Write("\nNo errors caught!")
    else:
        Write("\nWarnings and errors encountered:", 3)
        for i in errorlist:
            Write(i, 3)
    Write(message, 3)
    sys.exit(error)
    return

### Defining command line callbacks ###
def tabs_callback(option, opt, value, parser):
    global tab_length
    try:
        tab_length = int(parser.rargs[0])
        parser.rargs.remove(str(tab_length))
    except:
        tab_length = 4
    return



### Defining the command line parser ###
parser = optparse.OptionParser(description = help_string,
                               add_help_option = True,
                               prog = exe_name)

## For future conversion to argparse, if possible.
#parser.add_option("input",
#                  action = "store",
#                  nargs = "1", 
#                  help = "usage: TIconvert.py 'C:\Calc\etc\THEGAME.8xp'\n" \
#                         "The path of the 8xp or text file you wish to convert.  " \
#                         "This argument is mandatory.")
#
#parser.add_option("-p", "--path",
#                  action = "store",
#                  nargs = 1,
#                  help = "usage: TIconvert.py -p 'C:\Calc\\random\'\n" \
#                         "Changes the directory where the resulting file will be " \
#                         "written to.  Defaults to the directory where the original " \
#                         "is stored.",
#                    dest = "target_path")
parser.add_option("-l", "--log",
                  action = "store_true",
                  default = False, 
                  help = "usage: TIconvert.py -l\n" \
                         "If you chose to use this argument, this program creates a " \
                         "log file where the current program is stored.  This option " \
                         "is provided for debugging purposes only.")
parser.add_option("-c", "--colon", "--colons",
                  action = "store_true",
                  default = False,
                  help = "usage: TIconvert.py -c\n" \
                         "Use this if you want your text files to start with the colon at " \
                         "the start of each line.")
parser.add_option("-v", "--verbose",
                  action = "store_true",
                  default = False,
                  help = "usage: TIconvert.py -v\n" \
                         "Prints loads of info to the console.  Somewhat useful for " \
                         "debugging, although enabling logging would be a better idea.  "\
                         "Enable this if you want visual feedback that something has happened.")
parser.add_option("-t", "--tabs", "--tab",
                  action = "callback",
                  callback = tabs_callback,
                  help = "usage: TIconvert.py -t 4\n" \
                         "If this is enabled, it indents the code correctly when converting " \
                         "from 8xp to txt.  The number indicates how many spaces each tab " \
                         "should be.")


### Reading input from command line ###
# The parser may return some errors due to erronous user input.
# It quits automatically (I think?)

(options, arg) = parser.parse_args()




### Transforming input into friendlier data ###
Write("Calculating target destination...")

if arg == []:
    parser.parse_args(["--help"])

input_path = arg[0]



# Find root directory:
temp = input_path.rfind("\\")
root_path = input_path[0:temp]

# Find target path
input_name = input_path[temp:]
if "." in input_name:
    target_name = input_name[0:input_name.index(".")]
else:
    target_name = input_name
target_name += "_out.txt"
target_path = root_path + target_name


# Creating log file
is_logging_enabled = options.log
if is_logging_enabled:
    log_path = root_path + "\\TIConvert_log.txt"


# Quick error check:
DoesFileExist(input_path)
DoesFileExist(target_path)

if is_logging_enabled:
    log_file = open(log_path, "a")
    log_file.write(time.strftime("\n\n%Y-%m-%d %H:%M:%S\n"))   # Timestamp

Write("Input file:  " + input_path)
Write("Target name: " + target_path)
if is_logging_enabled:
    Write("Log file:    " + log_path)



### Preparing buffers ###
Write("Preparing buffers...")
input_file = open(input_path, "rb")  # read-only, binary
try:
    input_buffer = input_file.read() # Copy the entire contents of the file
except:
    QuitProgram("Could not accept input file -- possibly not enough memory?",2)
    
input_file.close()                   # Don't want to leave unclosable files hanging around.
output_buffer = ""

for char in input_buffer:
    output_buffer += NumHexToStr(char)
    

### Validating data ###
errorcount = 0

if output_buffer[0:22] != signature:
    errorcount += 1
    Write("WARNING: The file doesn't start with the correct signature - is it corrupted?", 2)

if (len(output_buffer) / 2 - 57) != ReturnTwoBytes(53, output_buffer):
    errorcount += 1
    Write("WARNING: The filesixe given in the header doesn't match the actual filesize.", 2)

variable_type = ReturnByte(59, output_buffer)
if variable_type in r_AllowedTypes:
    Write("Input file appears to be " + d_VariableTypes[variable_type] + ".")
    if variable_type > 7:
        errorcount != 1
        Write("WARNING: Detokenizing this file may return garbage.", 2)
else:
    errorcount != 1
    Write("WARNING: The file doesn't appear to be a program.", 2)


if errorcount:
    Write("Errors detected, proceeding anyways.", 2)
else:
    Write("No errors detected.")










### Setting up data ###
Write("Parsing preliminary data...")

offset = 74                             # First 74 bytes are header.
data_length = len(output_buffer) // 2 - offset

tab = ""
for i in range(tab_length):
    tab += " "

output_file = open(target_path, "w")    # Will overwrite any files with the same name
                                        # Change 2nd arg to "a" for appending.


previous_line = ""
current_line = ""

previous_depth = 0
current_depth = 0
next_depth = 0
normal_depth = 0

count = 0
byte1 = NumHexToStr(ReturnByte(offset, output_buffer))

if byte1 != "3A":
    Write("WARNING: This file doesn't start with a period, may not be Axe.", 2)

is_comment = True

prevcount = -1


### The actual tokenizing part ###
Write("Starting detokenizing...")
while True:
    if previous_line == None:
        break
    if options.colon:
        output_file.write(":")
    if (previous_depth > 0):
        for i in range(previous_depth):
            output_file.write(tab)
            
    #output_file.write(previous_line)
    output_file.write(previous_line)

    # The great shuffle
    previous_depth = current_depth
    current_depth = next_depth
    previous_line = current_line
    current_line = ""
    byte1 = ""

    try:
        if count >= (data_length-2):
            current_line = None
        else:
            check_tokens = True
            while byte1 != "3F":
                if count >= (data_length - 2):
                    raise
                byte1 = NumHexToStr(ReturnByte(count + offset, output_buffer))
                try:
                    byte2 = NumHexToStr(ReturnByte(count + offset + 1, output_buffer))
                except:
                    byte2 = "3F"
                if byte1 in r_TwoByteTokens:
                    current_line += d_AxeTokensMain[byte1][byte2]
                    count += 1  # Skip one.
                else:
                    
                    if byte1 == "3E":
                        check_tokens = True
                        
                    if byte1 == "3A":
                        is_comment = not(is_comment)

                    if check_tokens:
                        if byte1 in r_AxePush:
                            next_depth += 1
                            check_tokens = False
                        if byte1 in r_AxeBothP:
                            current_depth -= 1
                            check_tokens = False    
                        if byte1 in r_AxeNoP:
                            check_tokens = False
                        if byte1 in r_AxePop:
                            current_depth -= 1
                            next_depth -= 1
                            check_tokens = False
                                
                    current_line += d_AxeTokensMain[byte1]
                count += 1
    except:
        pass

    if 0 == (count % 50):
        # The percent thing here is slightly inaccurate, but it's not like anyone cares.
        Write("Approximately " + str(int(count / data_length * 100)) + "% done.")
        if count == prevcount:
            Write("ERROR: Unable to detokenize file - may be an assembly program.", 2)
            QuitProgram("Terminating detokenization...", 2)
        prevcount = count

        

QuitProgram("Finished detokenizing!\n")
























